#ident	"$Header: /var/cvsroot/htdb/src/lib/xml.c,v 1.6 2004/07/30 04:46:10 del Exp $"

#include	"libhtdb.h"

#if (defined(HAVE_LIBXML2) && defined(USE_LIBXML2))

#undef	DEBUG

/*
	prefix = "foo", name = "bar"	
	shortname == 0 ? 
		foo, bar  -> foo->bar[1]
	shortname == 1 ?
		foo, bar  -> foo->bar

	sets foo->bar->numResults as a side effect, ONLY when shortName is 0.
*/

static char	*htdb_parseXML_getdynname(char *prefix, char *name, int shortName)
{
	int cnt;
	static char buf[BUF_LIMIT_BIG + 1];


	if ( !prefix ) {
		/* dealing with the top level node */
		if ( shortName )
			return NULL;
		else
			snprintf(buf, BUF_LIMIT_BIG, name);
	} else {
		snprintf(buf, BUF_LIMIT_BIG, "%s->%s->numResults", prefix, name);
		cnt = htdb_getint(buf);

		if ( shortName ) {
			if ( cnt )
				return NULL;
			else
				snprintf(buf, BUF_LIMIT_BIG, "%s->%s", prefix, name);
		} else { 
			htdb_setint(buf, ++cnt);
			snprintf(buf, BUF_LIMIT_BIG, "%s->%s[%d]", prefix, name, cnt);
		}
	} 	
	
	return strdup(buf);
}

static void	htdb_parseXML_addField(char *prefix, char *name)
{
	int field_cnt;
	char *pref;

	field_cnt = htdb_getptrint(prefix, "numFields");
	field_cnt++;

	/*
			store field name in:
					prefix->field[]
	 */
	pref = htdb_getptrdyname(prefix, "field");
	htdb_setarrayval(pref, field_cnt, name);
#ifdef	DEBUG
	printf("setting %s[%d] to '%s'<br>\n", pref, field_cnt, name);
#endif
	free(pref);
	
	htdb_setptrint(prefix, "numFields", field_cnt);
}

static int	htdb_parseXML_node(char *prefix,  xmlNodePtr node, bool_t isRoot)
{
	int i;
	xmlAttrPtr attrib;
	xmlNodePtr walk = node;
	xmlChar *content;
	char *newprefix, *attribname;

	while ( walk ) {
		if ( !walk->name || xmlNodeIsText(walk) ) {
			walk = walk->next;		
			continue;
		}

		if (BOOL_TRUE == isRoot && prefix && (char *)walk->name && *(char *)walk->name &&
			!*htdb_getptrval(prefix, "root")) {
			htdb_setptrval(prefix, "root", (char *)walk->name);
		}

		/* this bit of complicated shit-code is so both prefix->blah and prefix[1]->blah get set */
		for(i=1; i >= 0; i--) {
			newprefix = htdb_parseXML_getdynname(prefix, (char *)walk->name, i);

			if ( !newprefix ) /* most likely couldn't get a short name for it */
				continue;

			content = xmlNodeGetContent(walk);	
			if ( content && *content ) {
				htdb_setval(newprefix, content);
#ifdef	DEBUG
				printf ("Setting %s<br>\n", newprefix);
#endif
				if ( prefix && i == 0 ) 	
					htdb_parseXML_addField(prefix, (char *) walk->name);
			}

			attrib = walk->properties;
			while(attrib) {
				attribname = htdb_parseXML_getdynname(newprefix, (char *) attrib->name, 1);
				content = xmlNodeGetContent(attrib->children);
				htdb_setval(attribname, content);			
#ifdef	DEBUG
				printf ("Setting %s<br>\n", attribname);
#endif
				htdb_parseXML_addField(newprefix, (char *) attrib->name);
				attrib = attrib->next;
				free(attribname);
			}		

			if ( walk->children ) 
				htdb_parseXML_node(newprefix, walk->children, BOOL_FALSE);

			free(newprefix);
		}

		walk = walk->next;
	}
	return 0;
}

/* 
	parse the XML in the given buffer to an htdb-friendly representation.

	if prefix is null, use the name of the first top level xml node:
		------
		<xbel version="1.0" folded="no" id="139597352">
			<title>My Bookmarks</title>
			 <info>
			 ....
		-----
		transforms to xbel->version = 1.0, 
					  xbel->title = "my Bookmarks",
		and on and on.
 */
bool_t		htdb_parseXML(char *prefix,  char *buf) 
{
	xmlNodePtr root;
#ifdef	DEBUG
	log_fatal("htdb_parseXML", "++");
#endif
	xmlInitParser();

	htdb_setptrval(prefix, "raw", buf);
	htdb_setptrval(prefix, "error", "");
	htdb_setptrval(prefix, "root", "");

	{
		xmlDocPtr doc = xmlParseDoc(buf);
		if ( !doc ) {
			htdb_setptrval(prefix, "error", "htdb_parseXML() document could not be parsed");
			return BOOL_FALSE;
		}

		root = xmlDocGetRootElement(doc);
		htdb_parseXML_node(prefix, root, BOOL_TRUE);

		xmlFreeDoc(doc);
	}
	xmlCleanupParser();

	if (NULL == prefix)
		return BOOL_TRUE;

	if (!*htdb_getptrval(prefix, "root")) {
		htdb_setptrval(prefix, "error", "htdb_parseXML() could not find namespace");
		return BOOL_FALSE;
	} else
		return BOOL_TRUE;
}

#else

bool_t		htdb_parseXML(char *prefix,  char *buf) 
{
	return BOOL_FALSE;
}

#endif
